"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const node_assert_1 = require("node:assert");
const index_js_1 = require("../../opLifecycle/index.js");
describe("BatchManager", () => {
    const hardLimit = 950 * 1024;
    const smallMessageSize = 10;
    const defaultOptions = {
        hardLimit,
        canRebase: true,
    };
    const generateStringOfSize = (sizeInBytes) => "0".repeat(sizeInBytes);
    const smallMessage = () => ({
        serializedOp: generateStringOfSize(smallMessageSize),
        referenceSequenceNumber: 0,
    });
    it("BatchManager: 'infinity' hard limit allows everything", () => {
        const message = {
            serializedOp: generateStringOfSize(1024),
            referenceSequenceNumber: 0,
        };
        const batchManager = new index_js_1.BatchManager({
            ...defaultOptions,
            hardLimit: Number.POSITIVE_INFINITY,
        });
        for (let i = 1; i <= 10; i++) {
            node_assert_1.strict.equal(batchManager.push(message, /* reentrant */ false), true);
            node_assert_1.strict.equal(batchManager.length, i);
        }
    });
    for (const includeBatchId of [true, false])
        it(`Batch metadata is set correctly [with${includeBatchId ? "" : "out"} batchId]`, () => {
            const batchManager = new index_js_1.BatchManager(defaultOptions);
            const batchId = includeBatchId ? "BATCH_ID" : undefined;
            node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
            /* reentrant */ false), true);
            node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
            /* reentrant */ false), true);
            node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
            /* reentrant */ false), true);
            const batch = batchManager.popBatch(batchId);
            node_assert_1.strict.deepEqual(batch.messages.map((m) => m.metadata), [
                { batch: true, ...(includeBatchId ? { batchId } : undefined) }, // batchId propertly should be omitted (v. set to undefined) if not provided
                undefined, // metadata not touched for intermediate messages
                { batch: false },
            ]);
            node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
            /* reentrant */ false), true);
            const singleOpBatch = batchManager.popBatch(batchId);
            node_assert_1.strict.deepEqual(singleOpBatch.messages.map((m) => m.metadata), [
                includeBatchId ? { batchId } : undefined, // batchId propertly should be omitted (v. set to undefined) if not provided
            ]);
        });
    it("BatchId Format", () => {
        const clientId = "3627a2a9-963f-4e3b-a4d2-a31b1267ef29";
        const batchStartCsn = 123;
        const batchId = (0, index_js_1.generateBatchId)(clientId, batchStartCsn);
        const serialized = JSON.stringify({ batchId });
        node_assert_1.strict.equal(serialized, `{"batchId":"3627a2a9-963f-4e3b-a4d2-a31b1267ef29_[123]"}`);
    });
    it("Batch content size is tracked correctly", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        node_assert_1.strict.equal(batchManager.push(smallMessage(), /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * batchManager.length);
        node_assert_1.strict.equal(batchManager.push(smallMessage(), /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * batchManager.length);
        node_assert_1.strict.equal(batchManager.push(smallMessage(), /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * batchManager.length);
    });
    it("Batch reference sequence number maps to the last message", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.sequenceNumbers.referenceSequenceNumber, 2);
    });
    const convertToOutboundBatch = (batch) => ({
        ...batch,
        messages: batch.messages.map((message) => ({
            ...message,
            contents: message.serializedOp,
            serializedOp: undefined,
        })),
    });
    it("Batch size estimates", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        batchManager.push(smallMessage(), /* reentrant */ false);
        // 10 bytes of content + 200 bytes overhead
        node_assert_1.strict.equal((0, index_js_1.estimateSocketSize)(convertToOutboundBatch(batchManager.popBatch())), 210);
        for (let i = 0; i < 10; i++) {
            batchManager.push(smallMessage(), /* reentrant */ false);
        }
        // (10 bytes of content + 200 bytes overhead) x 10
        node_assert_1.strict.equal((0, index_js_1.estimateSocketSize)(convertToOutboundBatch(batchManager.popBatch())), 2100);
        batchManager.push(smallMessage(), /* reentrant */ false);
        for (let i = 0; i < 9; i++) {
            batchManager.push({
                serializedOp: "",
                referenceSequenceNumber: 0,
            }, 
            /* reentrant */ false); // empty op
        }
        // 10 bytes of content + 200 bytes overhead x 10
        node_assert_1.strict.equal((0, index_js_1.estimateSocketSize)(convertToOutboundBatch(batchManager.popBatch())), 2010);
    });
    it("Batch op reentry state preserved during its lifetime", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.popBatch().hasReentrantOps, false);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ true, 
        /* currentClientSequenceNumber */ undefined), true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.popBatch().hasReentrantOps, true);
        node_assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        node_assert_1.strict.equal(batchManager.popBatch().hasReentrantOps, false);
    });
    it("should rollback to checkpoint correctly", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        // Push initial messages
        batchManager.push(smallMessage(), /* reentrant */ false);
        batchManager.push(smallMessage(), /* reentrant */ false);
        // Create checkpoint
        const checkpoint = batchManager.checkpoint();
        // Push more messages
        batchManager.push(smallMessage(), /* reentrant */ false);
        batchManager.push(smallMessage(), /* reentrant */ false);
        // Rollback to checkpoint
        checkpoint.rollback((message) => {
            // Process rollback message (no-op in this test)
        });
        // Verify state after rollback
        node_assert_1.strict.equal(batchManager.length, 2);
        node_assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * 2);
    });
    it("should handle rollback with no additional messages", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        // Push initial messages
        batchManager.push(smallMessage(), /* reentrant */ false);
        // Create checkpoint
        const checkpoint = batchManager.checkpoint();
        // Rollback to checkpoint without pushing more messages
        checkpoint.rollback((message) => {
            // Process rollback message (no-op in this test)
        });
        // Verify state after rollback
        node_assert_1.strict.equal(batchManager.length, 1);
        node_assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize);
    });
    it("should throw error if ops are generated during rollback", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        // Push initial messages
        batchManager.push(smallMessage(), /* reentrant */ false);
        // Create checkpoint
        const checkpoint = batchManager.checkpoint();
        // Push more messages
        batchManager.push(smallMessage(), /* reentrant */ false);
        // Attempt rollback and generate ops during rollback
        node_assert_1.strict.throws(() => {
            checkpoint.rollback((message) => {
                // Generate ops during rollback
                batchManager.push(smallMessage(), /* reentrant */ false);
            });
        }, /Error: Ops generated during rollback/);
    });
    it("Popping the batch then rolling is not allowed", () => {
        const batchManager = new index_js_1.BatchManager(defaultOptions);
        batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false);
        const checkpoint = batchManager.checkpoint();
        batchManager.popBatch();
        // Attempt rollback and generate ops during rollback
        node_assert_1.strict.throws(() => {
            checkpoint.rollback((message) => {
                // Generate ops during rollback
                batchManager.push(smallMessage(), /* reentrant */ false);
            });
        }, /Error: Ops generated during rollback/);
    });
});
//# sourceMappingURL=batchManager.spec.js.map